from mlgame.utils.enum import get_ai_name
from .gameMode import GameMode
from .maze_wall import *
from .sound_controller import SoundController
from mlgame.game.paia_game import GameResultState
from .tilemap import Map


class MoveMazeMode(GameMode):
    def __init__(self, user_num: int, maze_no, time, sensor, sound_controller):
        super(MoveMazeMode, self).__init__()
        '''load map data'''
        self.user_num = user_num
        self.maze_id = maze_no - 1
        self.map_file = "move_map_" + str(maze_no) + ".json"
        self.load_data()

        '''group of sprites'''
        self.worlds = []
        self.cars = pygame.sprite.Group()
        self.walls = pygame.sprite.Group()
        self.slant_walls = pygame.sprite.Group()
        self.wall_for_update = pygame.sprite.Group()
        self.all_points = pygame.sprite.Group()  # Group inclouding end point, check points,etc.

        '''data set'''

        self.wall_info = []
        self.wall_vertices_for_Box2D = []
        self.car_info = []
        self.ranked_user = []  # pygame.sprite car
        self.result = []
        self.eliminated_user = []
        self.user_check_points = []

        self.game_end_time = time  # int, decide how many frames the game will end even some users don't finish game
        pygame.font.init()
        self.state = GameResultState.UN_PASSED
        self.is_end = False
        self.sensor_num = sensor
        self.x = 0
        self._init_world(user_num)
        self.new()
        # '''sound'''
        # self.sound_controller = SoundController(sound_controller)

    def new(self):
        # initialize all variables and do all setup for a new game
        self.get_wall_info_v(2)
        self.get_wall_info_h(3)
        self.get_wall_info_h(1)
        self.get_wall_info_v(1)
        self.get_wall_info_v(4)
        self.get_wall_info_h(5)
        self.load_map_object()
        for wall_vertices in self.wall_vertices_for_Box2D:
            if wall_vertices["type"] == 1:
                for world in self.worlds:
                    wall = Wall(self, wall_vertices["vertices"], world)
                    if self.worlds.index(world) == 0:
                        self.walls.add(wall)
            elif wall_vertices["type"] == 2:
                for world in self.worlds:
                    wall = VerticalMoveWall(self, wall_vertices["vertices"], world, 4, 5)
                    self.wall_for_update.add(wall)
                    if self.worlds.index(world) == 0:
                        self.walls.add(wall)
            elif wall_vertices["type"] == 3:
                for world in self.worlds:
                    wall = HorizontalMoveWall(self, wall_vertices["vertices"], world, 3, 6)
                    self.wall_for_update.add(wall)
                    if self.worlds.index(world) == 0:
                        self.walls.add(wall)
            elif wall_vertices["type"] == 4:
                for world in self.worlds:
                    wall = VerticalMoveWall(self, wall_vertices["vertices"], world, 3, -5)
                    self.wall_for_update.add(wall)
                    if self.worlds.index(world) == 0:
                        self.walls.add(wall)
            elif wall_vertices["type"] == 5:
                for world in self.worlds:
                    wall = HorizontalMoveWall(self, wall_vertices["vertices"], world, 5, -7)
                    self.wall_for_update.add(wall)
                    if self.worlds.index(world) == 0:
                        self.walls.add(wall)

        self.pygame_point = [0, 0]

    def update_sprite(self, command):
        '''update the model of game,call this fuction per frame'''
        self.car_info.clear()
        self.frame += 1
        self.handle_event()
        self._is_game_end()
        self.command = command
        self.wall_for_update.update()
        self.wall_info.clear()
        for wall in self.walls:
            vertices = [(wall.body.transform * v) for v in wall.box.shape.vertices]
            self.wall_info.append([vertices[0], vertices[1]])
            self.wall_info.append([vertices[2], vertices[1]])
            self.wall_info.append([vertices[3], vertices[0]])
            self.wall_info.append([vertices[2], vertices[3]])
        for wall in self.slant_walls:
            vertices = [(wall.body.transform * v) for v in wall.box.shape.vertices]
            self.wall_info.append([vertices[0], vertices[1]])
            self.wall_info.append([vertices[1], vertices[2]])
            self.wall_info.append([vertices[2], vertices[0]])
        for car in self.cars:
            car.update(command[get_ai_name(car.car_no)])
            car.rect.center = self.trnsfer_box2d_to_pygame(car.body.position)
            self.car_info.append(car.get_info())
            car.detect_distance(self.frame, self.wall_info)

        for point in self.all_points:
            point.rect.x, point.rect.y = self.trnsfer_box2d_to_pygame((point.x, point.y))
        self.all_points.update()
        for world in self.worlds:
            world.Step(TIME_STEP, 10, 10)
            world.ClearForces()
        if self.is_end:
            self.running = False

    def load_data(self):
        map_folder = path.join(path.dirname(__file__), "map")
        try:
            self.map = Map(path.join(map_folder, self.map_file))
        except Exception:
            print(f"File '{self.map_file}' is not found.We will load first map for you.")
            self.map_file = "move_map_1.json"
            self.map = Map(path.join(map_folder, self.map_file))

        # print(self.map.data)

    def _init_world(self, user_no: int):
        for i in range(user_no):
            world = Box2D.b2.world(gravity=(0, 0), doSleep=True, CollideConnected=False)
            self.worlds.append(world)

    def _is_game_end(self):
        if self.frame >= self.game_end_time or len(self.eliminated_user) == len(self.cars):
            for car in self.cars:
                if car not in self.eliminated_user and car.is_running:
                    car.end_frame = self.frame
                    self.eliminated_user.append(car)
                    self.user_check_points.append(car.check_point)
                    car.is_running = False
                    car.status = "GAME_OVER"
            self.is_end = True
            self.ranked_user = self.rank()
            self._print_result()
            self.status = "END"

    def draw_grid(self):
        for x in range(TILE_LEFTTOP[0], TILE_WIDTH + TILE_LEFTTOP[0], TILESIZE):
            pygame.draw.line(self.screen, GREY, (x, TILE_LEFTTOP[1]), (x, TILE_HEIGHT + TILE_LEFTTOP[1]))
        for y in range(TILE_LEFTTOP[1], TILE_HEIGHT + TILE_LEFTTOP[1], TILESIZE):
            pygame.draw.line(self.screen, GREY, (TILE_LEFTTOP[0], y), (TILE_WIDTH + TILE_LEFTTOP[0], y))



import sys
import time

from mlgame.argument.cmd_argument import parse_cmd_and_get_arg_obj
from mlgame.argument.tool import revise_ai_clients
from mlgame.core.process import create_process_of_az_uploader_and_start
from mlgame.executor.game import GameExecutor
from mlgame.executor.manual import GameManualExecutor
from mlgame.game.paia_game import get_paia_game_obj
from mlgame.utils.logger import logger
from mlgame.view.audio_model import MusicInitSchema, SoundInitSchema
from mlgame.view.sound_controller import SoundController

if __name__ == "__main__":
    import os

    os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "hide"

    # 1. parse command line
    arg_obj = parse_cmd_and_get_arg_obj(sys.argv[1:])

    # 2. get parsed_game_params
    from mlgame.argument.game_argument import GameConfig

    game_config = GameConfig(arg_obj.game_folder.__str__())
    parsed_game_params = game_config.parse_game_params(arg_obj.game_params)
    # TODO use group AI
    ai_clients = revise_ai_clients(arg_obj.group_ai, game_config.user_num_config)
    user_num = len(ai_clients)
    game = get_paia_game_obj(game_config.game_cls, parsed_game_params, user_num,group_ai_list=ai_clients)

    # TODO refactor , process should not be init here   
    ai_process = []
    ws_proc = None
    record_proc = None
    az_upload_proc = None

    logger.info("===========Game is started ===========")
    from mlgame.core.communication import GameCommManager
    from mlgame.core.process import (
        create_process_of_ai_clients_and_start,
        create_process_of_ws_and_start,
        create_process_of_recorder_and_start,
        terminate,
    )
    from mlgame.view.view import PygameView, DummyPygameView

    game_comm = GameCommManager()
    try:
        if arg_obj.is_debug:
            logger.remove()
            logger.add(sys.stdout, level="DEBUG",format="<green>{time:YYYY-MM-DD HH:mm:ss.SSSSS}</green> | <level>{level: <8}</level> | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>")
            logger.add(
                "debug.log",
                level="DEBUG",
                rotation="10 MB",  # Rotate after the log file reaches 10 MB
                retention=1,  # Keep only the most recent rotated log file
                compression=None,  # Do not compress the log file
            )
            pass

        if arg_obj.ws_url:
            # prepare transmitter for game executor
            ws_proc = create_process_of_ws_and_start(game_comm, str(arg_obj.ws_url))

        if arg_obj.record_folder:
            # prepare transmitter for game executor
            record_proc = create_process_of_recorder_and_start(
                game_comm, arg_obj.record_folder, arg_obj.progress_frame_frequency
            )
        if arg_obj.az_upload_url:
            # prepare transmitter for game executor
            az_upload_proc = create_process_of_az_uploader_and_start(
                game_comm, arg_obj.az_upload_url
            )
        # 4. prepare ai_clients , create pipe, start ai_client process
        if arg_obj.is_manual:
            # only play in local and manual mode and will show view
            # TODO to deprecated and use ml_play_manual.py to alternate
            game_view = PygameView(
                game.get_scene_init_data(),
                caption=f"PAIA Game: {game_config.game_name} v{game_config.game_version}",
                icon=game_config.logo
            )
            game_executor = GameManualExecutor(
                game,
                game_view,
                game_comm,
                fps=arg_obj.fps,
                one_shot_mode=arg_obj.one_shot_mode,
            )

        else:
            init_data = game.get_scene_init_data()

            # set sound
            sound_controller = SoundController(
                is_sound_on=arg_obj.is_sound_on,
                music_objs=[
                    MusicInitSchema(**obj) for obj in init_data.get("musics", [])
                ],
                sound_objs=[
                    SoundInitSchema(**obj) for obj in init_data.get("sounds", [])
                ],
            )
            # play game with ai_clients
            if arg_obj.no_display:
                logger.warning("Game will not be displayed.")
                game_view = DummyPygameView(game.get_scene_init_data())
            else:
                game_view = PygameView(
                    game.get_scene_init_data(),
                    caption=f"PAIA Game: {game_config.game_name} v{game_config.game_version}",
                    icon=game_config.logo,
                    sound_controller=sound_controller,
                )
            ai_process = create_process_of_ai_clients_and_start(
                game_comm=game_comm,
                ai_clients=ai_clients,
                game_params=parsed_game_params,
            )

            # 5. run game in main process


            game_executor = GameExecutor(
                game,
                game_comm,
                game_view,
                # sound_controller=sound_controller,
                fps=arg_obj.fps,
                one_shot_mode=arg_obj.one_shot_mode,
                no_display=arg_obj.no_display,
                output_folder=arg_obj.output_folder,
            )

        time.sleep(0.1)
        game_executor.run()

    except Exception as e:
        # finally
        logger.exception(e)
        pass
    finally:
        # TODO refactor , process should not be init here
        terminate(game_comm, ai_process, ws_proc, record_proc,az_upload_proc)
    logger.info("===========All process is terminated ===========")
    pass



class MLPlay:
    def __init__(self, ai_name,*args,**kwargs):
        self.player_no = ai_name
        self.r_sensor_value = 0
        self.l_sensor_value = 0
        self.f_sensor_value = 0
        self.control_list = {"left_PWM" : 0, "right_PWM" : 0}
        # print("Initial ml script")
        print(kwargs)

    def update(self, scene_info: dict, *args, **kwargs):
        """
        Generate the command according to the received scene information
        """
        if scene_info["status"] != "GAME_ALIVE":
            return "RESET"
        self.r_sensor_value = scene_info["R_sensor"]
        self.l_sensor_value = scene_info["L_sensor"]
        self.f_sensor_value = scene_info["F_sensor"]
        if self.f_sensor_value >15:
            self.control_list["left_PWM"] = 100
            self.control_list["right_PWM"] = 100
        else:
            self.control_list["left_PWM"] = 0
            self.control_list["right_PWM"] = 0
        return self.control_list

    def reset(self):
        """
        Reset the status
        """
        # print("reset ml script")
        pass



import sys

import setuptools

# readme.md = github readme.md, 這裡可接受markdown寫法
# 如果沒有的話，需要自己打出介紹此專案的檔案，再讓程式知道
sys.path.append(r".")
import mlgame.version

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mlgame",  #
    version=mlgame.version.version,
    author="PAIA-Tech",
    author_email="service@paia-tech.com",
    description="A machine learning game framework based on Pygame",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/PAIA-Playful-AI-Arena/MLGame",
    packages=setuptools.find_packages(
        exclude=["tests"]
    ),
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.10.0, <3.13.0',
    include_package_data=True,
    keywords=["AI", "machine learning", 'game', 'framework'],

    install_requires=[
        # TODO refine here
        "pygame>=2.5.1,<3.0",
        'pandas==1.4.1',
        "pydantic>=2.0,<3.0",
        "websockets==10.2",
        "orjson>3.0,<4.0",
        "loguru>=0.7.2",
        "azure-storage-blob>=12.20"
    ]

)
